@@ -26,6 +26,7 @@ import (
 	"github.com/go-delve/delve/pkg/terminal"
 	"github.com/go-delve/delve/service/dap"
 	"github.com/go-delve/delve/service/dap/daptest"
+	"github.com/go-delve/delve/service/debugger"
 	"github.com/go-delve/delve/service/rpc2"
 	godap "github.com/google/go-dap"
 	"golang.org/x/tools/go/packages"
@@ -110,8 +111,13 @@ func TestBuild(t *testing.T) {
 
 	scan := bufio.NewScanner(stderr)
 	// wait for the debugger to start
-	scan.Scan()
-	t.Log(scan.Text())
+	for scan.Scan() {
+		text := scan.Text()
+		t.Log(text)
+		if strings.Contains(text, "API server pid = ") {
+			break
+		}
+	}
 	go func() {
 		for scan.Scan() {
 			t.Log(scan.Text())
@@ -281,8 +287,10 @@ func TestContinue(t *testing.T) {
 
 // TestChildProcessExitWhenNoDebugInfo verifies that the child process exits when dlv launch the binary without debug info
 func TestChildProcessExitWhenNoDebugInfo(t *testing.T) {
+	noDebugFlags := protest.LinkStrip
+	// -s doesn't strip symbols on Mac, use -w instead
 	if runtime.GOOS == "darwin" {
-		t.Skip("test skipped on darwin, see https://github.com/go-delve/delve/pull/2018 for details")
+		noDebugFlags = protest.LinkDisableDWARF
 	}
 
 	if _, err := exec.LookPath("ps"); err != nil {
@@ -292,12 +300,18 @@ func TestChildProcessExitWhenNoDebugInfo(t *testing.T) {
 	dlvbin, tmpdir := getDlvBin(t)
 	defer os.RemoveAll(tmpdir)
 
-	fix := protest.BuildFixture("http_server", protest.LinkStrip)
+	fix := protest.BuildFixture("http_server", noDebugFlags)
 
 	// dlv exec the binary file and expect error.
-	if _, err := exec.Command(dlvbin, "exec", fix.Path).CombinedOutput(); err == nil {
+	out, err := exec.Command(dlvbin, "exec", "--headless", "--log", fix.Path).CombinedOutput()
+	t.Log(string(out))
+	if err == nil {
 		t.Fatalf("Expected err when launching the binary without debug info, but got nil")
 	}
+	//  Test only for dlv's prefix of the error like "could not launch process: could not open debug info"
+	if !strings.Contains(string(out), "could not launch process") || !strings.Contains(string(out), debugger.NoDebugWarning) {
+		t.Fatalf("Expected logged error 'could not launch process: ... - %s'", debugger.NoDebugWarning)
+	}
 
 	// search the running process named fix.Name
 	cmd := exec.Command("ps", "-aux")
@@ -693,9 +707,64 @@ func TestDAPCmd(t *testing.T) {
 	cmd.Wait()
 }
 
-func newDAPRemoteClient(t *testing.T, addr string) *daptest.Client {
+func TestDAPCmdWithNoDebugBinary(t *testing.T) {
+	const listenAddr = "127.0.0.1:40579"
+
+	dlvbin, tmpdir := getDlvBin(t)
+	defer os.RemoveAll(tmpdir)
+
+	cmd := exec.Command(dlvbin, "dap", "--log", "--listen", listenAddr)
+	stdout, err := cmd.StdoutPipe()
+	assertNoError(err, t, "stdout pipe")
+	defer stdout.Close()
+	stderr, err := cmd.StderrPipe()
+	assertNoError(err, t, "stderr pipe")
+	defer stderr.Close()
+	assertNoError(cmd.Start(), t, "start dap instance")
+
+	scanOut := bufio.NewScanner(stdout)
+	scanErr := bufio.NewScanner(stderr)
+	// Wait for the debug server to start
+	scanOut.Scan()
+	listening := "DAP server listening at: " + listenAddr
+	if scanOut.Text() != listening {
+		cmd.Process.Kill() // release the port
+		t.Fatalf("Unexpected stdout:\ngot  %q\nwant %q", scanOut.Text(), listening)
+	}
+	go func() { // Capture logging
+		for scanErr.Scan() {
+			t.Log(scanErr.Text())
+		}
+	}()
+
+	// Exec the stripped debuggee and expect things to fail
+	noDebugFlags := protest.LinkStrip
+	// -s doesn't strip symbols on Mac, use -w instead
+	if runtime.GOOS == "darwin" {
+		noDebugFlags = protest.LinkDisableDWARF
+	}
+	fixture := protest.BuildFixture("increment", noDebugFlags)
+	go func() {
+		for scanOut.Scan() {
+			t.Errorf("Unexpected stdout: %s", scanOut.Text())
+		}
+	}()
+	client := daptest.NewClient(listenAddr)
+	client.LaunchRequest("exec", fixture.Path, false)
+	client.ExpectErrorResponse(t)
+	client.DisconnectRequest()
+	client.ExpectDisconnectResponse(t)
+	client.ExpectTerminatedEvent(t)
+	client.Close()
+	cmd.Wait()
+}
+
+func newDAPRemoteClient(t *testing.T, addr string, isDlvAttach bool, isMulti bool) *daptest.Client {
 	c := daptest.NewClient(addr)
 	c.AttachRequest(map[string]interface{}{"mode": "remote", "stopOnEntry": true})
+	if isDlvAttach || isMulti {
+		c.ExpectCapabilitiesEventSupportTerminateDebuggee(t)
+	}
 	c.ExpectInitializedEvent(t)
 	c.ExpectAttachResponse(t)
 	c.ConfigurationDoneRequest()
@@ -732,7 +801,7 @@ func TestRemoteDAPClient(t *testing.T) {
 		}
 	}()
 
-	client := newDAPRemoteClient(t, listenAddr)
+	client := newDAPRemoteClient(t, listenAddr, false, false)
 	client.ContinueRequest(1)
 	client.ExpectContinueResponse(t)
 	client.ExpectTerminatedEvent(t)
@@ -792,7 +861,7 @@ func TestRemoteDAPClientMulti(t *testing.T) {
 	dapclient0.ExpectErrorResponse(t)
 
 	// Client 1 connects and continues to main.main
-	dapclient := newDAPRemoteClient(t, listenAddr)
+	dapclient := newDAPRemoteClient(t, listenAddr, false, true)
 	dapclient.SetFunctionBreakpointsRequest([]godap.FunctionBreakpoint{{Name: "main.main"}})
 	dapclient.ExpectSetFunctionBreakpointsResponse(t)
 	dapclient.ContinueRequest(1)
@@ -802,14 +871,14 @@ func TestRemoteDAPClientMulti(t *testing.T) {
 	closeDAPRemoteMultiClient(t, dapclient, "halted")
 
 	// Client 2 reconnects at main.main and continues to process exit
-	dapclient2 := newDAPRemoteClient(t, listenAddr)
+	dapclient2 := newDAPRemoteClient(t, listenAddr, false, true)
 	dapclient2.CheckStopLocation(t, 1, "main.main", 5)
 	dapclient2.ContinueRequest(1)
 	dapclient2.ExpectContinueResponse(t)
 	dapclient2.ExpectTerminatedEvent(t)
 	closeDAPRemoteMultiClient(t, dapclient2, "exited")
 
-	// Attach to exited processs is an error
+	// Attach to exited processes is an error
 	dapclient3 := daptest.NewClient(listenAddr)
 	dapclient3.AttachRequest(map[string]interface{}{"mode": "remote", "stopOnEntry": true})
 	dapclient3.ExpectErrorResponseWith(t, dap.FailedToAttach, `Process \d+ has exited with status 0`, true)
@@ -862,7 +931,7 @@ func TestRemoteDAPClientAfterContinue(t *testing.T) {
 		}
 	}()
 
-	c := newDAPRemoteClient(t, listenAddr)
+	c := newDAPRemoteClient(t, listenAddr, false, true)
 	c.ContinueRequest(1)
 	c.ExpectContinueResponse(t)
 	c.DisconnectRequest()
@@ -871,7 +940,7 @@ func TestRemoteDAPClientAfterContinue(t *testing.T) {
 	c.ExpectTerminatedEvent(t)
 	c.Close()
 
-	c = newDAPRemoteClient(t, listenAddr)
+	c = newDAPRemoteClient(t, listenAddr, false, true)
 	c.DisconnectRequestWithKillOption(true)
 	c.ExpectOutputEventDetachingKill(t)
 	c.ExpectDisconnectResponse(t)
@@ -929,7 +998,7 @@ func TestTrace(t *testing.T) {
 	dlvbin, tmpdir := getDlvBin(t)
 	defer os.RemoveAll(tmpdir)
 
-	expected := []byte("> goroutine(1): main.foo(99, 9801) => (9900)\n")
+	expected := []byte("> goroutine(1): main.foo(99, 9801)\n>> goroutine(1): => (9900)\n")
 
 	fixtures := protest.FindFixturesDir()
 	cmd := exec.Command(dlvbin, "trace", "--output", filepath.Join(tmpdir, "__debug"), filepath.Join(fixtures, "issue573.go"), "foo")
@@ -950,6 +1019,38 @@ func TestTrace(t *testing.T) {
 	cmd.Wait()
 }
 
+func TestTraceMultipleGoroutines(t *testing.T) {
+	dlvbin, tmpdir := getDlvBin(t)
+	defer os.RemoveAll(tmpdir)
+
+	// TODO(derekparker) this test has to be a bit vague to avoid flakyness.
+	// I think a future improvement could be to use regexp captures to match the
+	// goroutine IDs at function entry and exit.
+	expected := []byte("main.callme(0, \"five\")\n")
+	expected2 := []byte("=> (0)\n")
+
+	fixtures := protest.FindFixturesDir()
+	cmd := exec.Command(dlvbin, "trace", "--output", filepath.Join(tmpdir, "__debug"), filepath.Join(fixtures, "goroutines-trace.go"), "callme")
+	rdr, err := cmd.StderrPipe()
+	assertNoError(err, t, "stderr pipe")
+	defer rdr.Close()
+
+	cmd.Dir = filepath.Join(fixtures, "buildtest")
+
+	assertNoError(cmd.Start(), t, "running trace")
+
+	output, err := ioutil.ReadAll(rdr)
+	assertNoError(err, t, "ReadAll")
+
+	if !bytes.Contains(output, expected) {
+		t.Fatalf("expected:\n%s\ngot:\n%s", string(expected), string(output))
+	}
+	if !bytes.Contains(output, expected2) {
+		t.Fatalf("expected:\n%s\ngot:\n%s", string(expected), string(output))
+	}
+	cmd.Wait()
+}
+
 func TestTracePid(t *testing.T) {
 	if runtime.GOOS == "linux" {
 		bs, _ := ioutil.ReadFile("/proc/sys/kernel/yama/ptrace_scope")
@@ -962,7 +1063,7 @@ func TestTracePid(t *testing.T) {
 	dlvbin, tmpdir := getDlvBin(t)
 	defer os.RemoveAll(tmpdir)
 
-	expected := []byte("goroutine(1): main.A()\n => ()\n")
+	expected := []byte("goroutine(1): main.A()\n>> goroutine(1): => ()\n")
 
 	// make process run
 	fix := protest.BuildFixture("issue2023", 0)
@@ -1082,6 +1183,67 @@ func TestTraceEBPF(t *testing.T) {
 	cmd.Wait()
 }
 
+func TestTraceEBPF2(t *testing.T) {
+	if os.Getenv("CI") == "true" {
+		t.Skip("cannot run test in CI, requires kernel compiled with btf support")
+	}
+	if runtime.GOOS != "linux" || runtime.GOARCH != "amd64" {
+		t.Skip("not implemented on non linux/amd64 systems")
+	}
+	if !goversion.VersionAfterOrEqual(runtime.Version(), 1, 16) {
+		t.Skip("requires at least Go 1.16 to run test")
+	}
+	usr, err := user.Current()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if usr.Uid != "0" {
+		t.Skip("test must be run as root")
+	}
+
+	dlvbin, tmpdir := getDlvBinEBPF(t)
+	defer os.RemoveAll(tmpdir)
+
+	expected := []byte(`> (1) main.callme(10)
+> (1) main.callme(9)
+> (1) main.callme(8)
+> (1) main.callme(7)
+> (1) main.callme(6)
+> (1) main.callme(5)
+> (1) main.callme(4)
+> (1) main.callme(3)
+> (1) main.callme(2)
+> (1) main.callme(1)
+> (1) main.callme(0)
+=> "100"
+=> "100"
+=> "100"
+=> "100"
+=> "100"
+=> "100"
+=> "100"
+=> "100"
+=> "100"
+=> "100"
+=> "100"`)
+
+	fixtures := protest.FindFixturesDir()
+	cmd := exec.Command(dlvbin, "trace", "--ebpf", "--output", filepath.Join(tmpdir, "__debug"), filepath.Join(fixtures, "ebpf_trace.go"), "main.callme")
+	rdr, err := cmd.StderrPipe()
+	assertNoError(err, t, "stderr pipe")
+	defer rdr.Close()
+
+	assertNoError(cmd.Start(), t, "running trace")
+
+	output, err := ioutil.ReadAll(rdr)
+	assertNoError(err, t, "ReadAll")
+
+	if !bytes.Contains(output, expected) {
+		t.Fatalf("expected:\n%s\ngot:\n%s", string(expected), string(output))
+	}
+	cmd.Wait()
+}
+
 func TestDlvTestChdir(t *testing.T) {
 	dlvbin, tmpdir := getDlvBin(t)
 	defer os.RemoveAll(tmpdir)
@@ -1118,10 +1280,6 @@ func TestVersion(t *testing.T) {
 }
 
 func TestStaticcheck(t *testing.T) {
-	if goversion.VersionAfterOrEqual(runtime.Version(), 1, 18) {
-		//TODO(aarzilli): remove this before version 1.8.0 is released
-		t.Skip("staticcheck does not currently support Go 1.18")
-	}
 	_, err := exec.LookPath("staticcheck")
 	if err != nil {
 		t.Skip("staticcheck not installed")
