@@ -36,10 +36,10 @@ const (
 // Target represents the process being debugged.
 type Target struct {
 	Process
-	RecordingManipulation
 
-	proc   ProcessInternal
-	recman RecordingManipulationInternal
+	proc         ProcessInternal
+	recman       RecordingManipulationInternal
+	continueOnce ContinueOnceFunc
 
 	pid int
 
@@ -51,10 +51,6 @@ type Target struct {
 	// CanDump is true if core dumping is supported.
 	CanDump bool
 
-	// KeepSteppingBreakpoints determines whether certain stop reasons (e.g. manual halts)
-	// will keep the stepping breakpoints instead of clearing them.
-	KeepSteppingBreakpoints KeepSteppingBreakpoints
-
 	// currentThread is the thread that will be used by next/step/stepout and to evaluate variables if no goroutine is selected.
 	currentThread Thread
 
@@ -63,7 +59,7 @@ type Target struct {
 	selectedGoroutine *G
 
 	// fncallForG stores a mapping of current active function calls.
-	fncallForG map[int]*callInjection
+	fncallForG map[int64]*callInjection
 
 	asyncPreemptChanged bool  // runtime/debug.asyncpreemptoff was changed
 	asyncPreemptOff     int64 // cached value of runtime/debug.asyncpreemptoff
@@ -83,6 +79,8 @@ type Target struct {
 	// can be given a unique address.
 	fakeMemoryRegistry    []*compositeMemory
 	fakeMemoryRegistryMap map[string]*compositeMemory
+
+	partOfGroup bool
 }
 
 type KeepSteppingBreakpoints uint8
@@ -149,6 +147,8 @@ const (
 	StopWatchpoint                     // The target process hit one or more watchpoints
 )
 
+type ContinueOnceFunc func([]ProcessInternal, *ContinueOnceContext) (trapthread Thread, stopReason StopReason, err error)
+
 // NewTargetConfig contains the configuration for a new Target object,
 type NewTargetConfig struct {
 	Path                string     // path of the main executable
@@ -156,6 +156,7 @@ type NewTargetConfig struct {
 	DisableAsyncPreempt bool       // Go 1.14 asynchronous preemption should be disabled
 	StopReason          StopReason // Initial stop reason
 	CanDump             bool       // Can create core dumps (must implement ProcessInternal.MemoryMap)
+	ContinueOnce        ContinueOnceFunc
 }
 
 // DisableAsyncPreemptEnv returns a process environment (like os.Environ)
@@ -193,11 +194,12 @@ func NewTarget(p ProcessInternal, pid int, currentThread Thread, cfg NewTargetCo
 	t := &Target{
 		Process:       p,
 		proc:          p,
-		fncallForG:    make(map[int]*callInjection),
+		fncallForG:    make(map[int64]*callInjection),
 		StopReason:    cfg.StopReason,
 		currentThread: currentThread,
 		CanDump:       cfg.CanDump,
 		pid:           pid,
+		continueOnce:  cfg.ContinueOnce,
 	}
 
 	if recman, ok := p.(RecordingManipulationInternal); ok {
@@ -205,7 +207,6 @@ func NewTarget(p ProcessInternal, pid int, currentThread Thread, cfg NewTargetCo
 	} else {
 		t.recman = &dummyRecordingManipulation{}
 	}
-	t.RecordingManipulation = t.recman
 
 	g, _ := GetG(currentThread)
 	t.selectedGoroutine = g
@@ -265,7 +266,7 @@ func (t *Target) Valid() (bool, error) {
 // Currently only non-recorded processes running on AMD64 support
 // function calls.
 func (t *Target) SupportsFunctionCalls() bool {
-	return t.Process.BinInfo().Arch.Name == "amd64"
+	return t.Process.BinInfo().Arch.Name == "amd64" || t.Process.BinInfo().Arch.Name == "arm64"
 }
 
 // ClearCaches clears internal caches that should not survive a restart.
@@ -278,12 +279,18 @@ func (t *Target) ClearCaches() {
 	}
 }
 
-// Restart will start the process over from the location specified by the "from" locspec.
+// Restart will start the process group over from the location specified by the "from" locspec.
 // This is only useful for recorded targets.
 // Restarting of a normal process happens at a higher level (debugger.Restart).
-func (t *Target) Restart(from string) error {
-	t.ClearCaches()
-	currentThread, err := t.recman.Restart(from)
+func (grp *TargetGroup) Restart(from string) error {
+	if len(grp.targets) != 1 {
+		panic("multiple targets not implemented")
+	}
+	for _, t := range grp.targets {
+		t.ClearCaches()
+	}
+	t := grp.Selected
+	currentThread, err := t.recman.Restart(grp.cctx, from)
 	if err != nil {
 		return err
 	}
@@ -330,11 +337,11 @@ func (p *Target) SwitchThread(tid int) error {
 	return fmt.Errorf("thread %d does not exist", tid)
 }
 
-// Detach will detach the target from the underylying process.
+// detach will detach the target from the underylying process.
 // This means the debugger will no longer receive events from the process
 // we were previously debugging.
 // If kill is true then the process will be killed when we detach.
-func (t *Target) Detach(kill bool) error {
+func (t *Target) detach(kill bool) error {
 	if !kill {
 		if t.asyncPreemptChanged {
 			setAsyncPreemptOff(t, t.asyncPreemptOff)
@@ -393,10 +400,10 @@ func (t *Target) createUnrecoveredPanicBreakpoint() {
 		panicpcs, err = FindFunctionLocation(t.Process, "runtime.fatalpanic", 0)
 	}
 	if err == nil {
-		bp, err := t.SetBreakpointWithID(unrecoveredPanicID, panicpcs[0])
+		bp, err := t.SetBreakpoint(unrecoveredPanicID, panicpcs[0], UserBreakpoint, nil)
 		if err == nil {
-			bp.Name = UnrecoveredPanic
-			bp.Variables = []string{"runtime.curg._panic.arg"}
+			bp.Logical.Name = UnrecoveredPanic
+			bp.Logical.Variables = []string{"runtime.curg._panic.arg"}
 		}
 	}
 }
@@ -405,9 +412,16 @@ func (t *Target) createUnrecoveredPanicBreakpoint() {
 func (t *Target) createFatalThrowBreakpoint() {
 	fatalpcs, err := FindFunctionLocation(t.Process, "runtime.throw", 0)
 	if err == nil {
-		bp, err := t.SetBreakpointWithID(fatalThrowID, fatalpcs[0])
+		bp, err := t.SetBreakpoint(fatalThrowID, fatalpcs[0], UserBreakpoint, nil)
 		if err == nil {
-			bp.Name = FatalThrow
+			bp.Logical.Name = FatalThrow
+		}
+	}
+	fatalpcs, err = FindFunctionLocation(t.Process, "runtime.fatal", 0)
+	if err == nil {
+		bp, err := t.SetBreakpoint(fatalThrowID, fatalpcs[0], UserBreakpoint, nil)
+		if err == nil {
+			bp.Logical.Name = FatalThrow
 		}
 	}
 }
@@ -422,6 +436,7 @@ func (t *Target) CurrentThread() Thread {
 type UProbeTraceResult struct {
 	FnAddr       int
 	GoroutineID  int
+	IsRet        bool
 	InputParams  []*Variable
 	ReturnParams []*Variable
 }
@@ -451,6 +466,7 @@ func (t *Target) GetBufferedTracepoints() []*UProbeTraceResult {
 		r := &UProbeTraceResult{}
 		r.FnAddr = tp.FnAddr
 		r.GoroutineID = tp.GoroutineID
+		r.IsRet = tp.IsRet
 		for _, ip := range tp.InputParams {
 			v := convertInputParamToVariable(ip)
 			r.InputParams = append(r.InputParams, v)
@@ -464,14 +480,23 @@ func (t *Target) GetBufferedTracepoints() []*UProbeTraceResult {
 	return results
 }
 
-// SetNextBreakpointID sets the breakpoint ID of the next breakpoint
-func (t *Target) SetNextBreakpointID(id int) {
-	t.Breakpoints().breakpointIDCounter = id
+// ResumeNotify specifies a channel that will be closed the next time
+// Continue finishes resuming the targets.
+func (t *TargetGroup) ResumeNotify(ch chan<- struct{}) {
+	t.cctx.ResumeChan = ch
+}
+
+// RequestManualStop attempts to stop all the processes' threads.
+func (t *TargetGroup) RequestManualStop() error {
+	t.cctx.StopMu.Lock()
+	defer t.cctx.StopMu.Unlock()
+	t.cctx.manualStopRequested = true
+	return t.Selected.proc.RequestManualStop(t.cctx)
 }
 
 const (
 	FakeAddressBase     = 0xbeef000000000000
-	fakeAddressUnresolv = 0xbeed000000000000 // this address never resloves to memory
+	fakeAddressUnresolv = 0xbeed000000000000 // this address never resolves to memory
 )
 
 // newCompositeMemory creates a new compositeMemory object and registers it.
@@ -487,7 +512,7 @@ func (t *Target) newCompositeMemory(mem MemoryReadWriter, regs op.DwarfRegisters
 		// this combination is guaranteed to be unique between resumes.
 		buf := new(strings.Builder)
 		fmt.Fprintf(buf, "%#x ", regs.CFA)
-		op.PrettyPrint(buf, descr.instr)
+		op.PrettyPrint(buf, descr.instr, t.BinInfo().Arch.RegnumToString)
 		key = buf.String()
 
 		if cmem := t.fakeMemoryRegistryMap[key]; cmem != nil {
@@ -598,4 +623,6 @@ func (*dummyRecordingManipulation) ClearCheckpoint(int) error { return ErrNotRec
 
 // Restart will always return an error in the native proc backend, only for
 // recorded traces.
-func (*dummyRecordingManipulation) Restart(string) (Thread, error) { return nil, ErrNotRecorded }
+func (*dummyRecordingManipulation) Restart(*ContinueOnceContext, string) (Thread, error) {
+	return nil, ErrNotRecorded
+}
