@@ -13,6 +13,7 @@ import (
 	"sort"
 	"strconv"
 	"strings"
+	"time"
 	"unsafe"
 
 	"github.com/go-delve/delve/pkg/dwarf/godwarf"
@@ -76,7 +77,7 @@ const (
 	VariableReturnArgument
 	// VariableFakeAddress means the address of this variable is either fake
 	// (i.e. the variable is partially or completely stored in a CPU register
-	// and doesn't have a real address) or possibly no longer availabe (because
+	// and doesn't have a real address) or possibly no longer available (because
 	// the variable is the return value of a function call and allocated on a
 	// frame that no longer exists)
 	VariableFakeAddress
@@ -195,7 +196,7 @@ const (
 // G represents a runtime G (goroutine) structure (at least the
 // fields that Delve is interested in).
 type G struct {
-	ID      int    // Goroutine ID
+	ID      int64  // Goroutine ID
 	PC      uint64 // PC of goroutine when it was parked.
 	SP      uint64 // SP of goroutine when it was parked.
 	BP      uint64 // BP of goroutine when it was parked (go >= 1.7).
@@ -311,7 +312,7 @@ func GoroutinesInfo(dbp *Target, start, count int) ([]*G, int, error) {
 	}
 
 	var (
-		threadg = map[int]*G{}
+		threadg = map[int64]*G{}
 		allg    []*G
 	)
 
@@ -366,7 +367,7 @@ func GoroutinesInfo(dbp *Target, start, count int) ([]*G, int, error) {
 
 // FindGoroutine returns a G struct representing the goroutine
 // specified by `gid`.
-func FindGoroutine(dbp *Target, gid int) (*G, error) {
+func FindGoroutine(dbp *Target, gid int64) (*G, error) {
 	if selg := dbp.SelectedGoroutine(); (gid == -1) || (selg != nil && selg.ID == gid) || (selg == nil && gid == 0) {
 		// Return the currently selected goroutine in the following circumstances:
 		//
@@ -739,11 +740,16 @@ func resolveTypedef(typ godwarf.Type) godwarf.Type {
 	}
 }
 
+var constantMaxInt64 = constant.MakeInt64(1<<63 - 1)
+
 func newConstant(val constant.Value, mem MemoryReadWriter) *Variable {
 	v := &Variable{Value: val, mem: mem, loaded: true}
 	switch val.Kind() {
 	case constant.Int:
 		v.Kind = reflect.Int
+		if constant.Sign(val) >= 0 && constant.Compare(val, token.GTR, constantMaxInt64) {
+			v.Kind = reflect.Uint64
+		}
 	case constant.Float:
 		v.Kind = reflect.Float64
 	case constant.Bool:
@@ -880,7 +886,17 @@ func (v *Variable) parseG() (*G, error) {
 		return n
 	}
 
-	id := loadInt64Maybe("goid")             // +rtype int64
+	loadUint64Maybe := func(name string) uint64 {
+		vv := v.loadFieldNamed(name)
+		if vv == nil {
+			unreadable = true
+			return 0
+		}
+		n, _ := constant.Uint64Val(vv.Value)
+		return n
+	}
+
+	id := loadUint64Maybe("goid")            // +rtype int64|uint64
 	gopc := loadInt64Maybe("gopc")           // +rtype uintptr
 	startpc := loadInt64Maybe("startpc")     // +rtype uintptr
 	waitSince := loadInt64Maybe("waitsince") // +rtype int64
@@ -890,11 +906,15 @@ func (v *Variable) parseG() (*G, error) {
 	}
 	var stackhi, stacklo uint64
 	if stackVar := v.loadFieldNamed("stack"); /* +rtype stack */ stackVar != nil {
-		if stackhiVar := stackVar.fieldVariable("hi"); /* +rtype uintptr */ stackhiVar != nil {
+		if stackhiVar := stackVar.fieldVariable("hi"); /* +rtype uintptr */ stackhiVar != nil && stackhiVar.Value != nil {
 			stackhi, _ = constant.Uint64Val(stackhiVar.Value)
+		} else {
+			unreadable = true
 		}
-		if stackloVar := stackVar.fieldVariable("lo"); /* +rtype uintptr */ stackloVar != nil {
+		if stackloVar := stackVar.fieldVariable("lo"); /* +rtype uintptr */ stackloVar != nil && stackloVar.Value != nil {
 			stacklo, _ = constant.Uint64Val(stackloVar.Value)
+		} else {
+			unreadable = true
 		}
 	}
 
@@ -909,7 +929,7 @@ func (v *Variable) parseG() (*G, error) {
 	v.Name = "runtime.curg"
 
 	g := &G{
-		ID:         int(id),
+		ID:         int64(id),
 		GoPC:       uint64(gopc),
 		StartPC:    uint64(startpc),
 		PC:         uint64(pc),
@@ -1304,6 +1324,9 @@ func (v *Variable) loadValueInternal(recurseLevel int, cfg LoadConfig) {
 				v.Children[i].loadValueInternal(recurseLevel+1, cfg)
 			}
 		}
+		if t.Name == "time.Time" {
+			v.formatTime()
+		}
 
 	case reflect.Interface:
 		v.loadInterface(recurseLevel, true, cfg)
@@ -2274,37 +2297,47 @@ func (v *Variable) registerVariableTypeConv(newtyp string) (*Variable, error) {
 		switch newtyp {
 		case "int8":
 			child = newConstant(constant.MakeInt64(int64(int8(v.reg.Bytes[i]))), v.mem)
+			child.Kind = reflect.Int8
 			n = 1
 		case "int16":
 			child = newConstant(constant.MakeInt64(int64(int16(binary.LittleEndian.Uint16(v.reg.Bytes[i:])))), v.mem)
+			child.Kind = reflect.Int16
 			n = 2
 		case "int32":
 			child = newConstant(constant.MakeInt64(int64(int32(binary.LittleEndian.Uint32(v.reg.Bytes[i:])))), v.mem)
+			child.Kind = reflect.Int32
 			n = 4
 		case "int64":
 			child = newConstant(constant.MakeInt64(int64(binary.LittleEndian.Uint64(v.reg.Bytes[i:]))), v.mem)
+			child.Kind = reflect.Int64
 			n = 8
 		case "uint8":
 			child = newConstant(constant.MakeUint64(uint64(v.reg.Bytes[i])), v.mem)
+			child.Kind = reflect.Uint8
 			n = 1
 		case "uint16":
 			child = newConstant(constant.MakeUint64(uint64(binary.LittleEndian.Uint16(v.reg.Bytes[i:]))), v.mem)
+			child.Kind = reflect.Uint16
 			n = 2
 		case "uint32":
 			child = newConstant(constant.MakeUint64(uint64(binary.LittleEndian.Uint32(v.reg.Bytes[i:]))), v.mem)
+			child.Kind = reflect.Uint32
 			n = 4
 		case "uint64":
 			child = newConstant(constant.MakeUint64(uint64(binary.LittleEndian.Uint64(v.reg.Bytes[i:]))), v.mem)
+			child.Kind = reflect.Uint64
 			n = 8
 		case "float32":
 			a := binary.LittleEndian.Uint32(v.reg.Bytes[i:])
 			x := *(*float32)(unsafe.Pointer(&a))
 			child = newConstant(constant.MakeFloat64(float64(x)), v.mem)
+			child.Kind = reflect.Float32
 			n = 4
 		case "float64":
 			a := binary.LittleEndian.Uint64(v.reg.Bytes[i:])
 			x := *(*float64)(unsafe.Pointer(&a))
 			child = newConstant(constant.MakeFloat64(x), v.mem)
+			child.Kind = reflect.Float64
 			n = 8
 		default:
 			if n == 0 {
@@ -2454,3 +2487,63 @@ type constantValuesByValue []constantValue
 func (v constantValuesByValue) Len() int               { return len(v) }
 func (v constantValuesByValue) Less(i int, j int) bool { return v[i].value < v[j].value }
 func (v constantValuesByValue) Swap(i int, j int)      { v[i], v[j] = v[j], v[i] }
+
+const (
+	timeTimeWallHasMonotonicBit uint64 = (1 << 63) // hasMonotonic bit of time.Time.wall
+
+	//lint:ignore ST1011 addSeconds is the name of the relevant function
+	maxAddSeconds time.Duration = (time.Duration(^uint64(0)>>1) / time.Second) * time.Second // maximum number of seconds that can be added with (time.Time).Add, measured in nanoseconds
+
+	wallNsecShift = 30 // size of the nanoseconds field of time.Time.wall
+
+	unixTimestampOfWallEpoch = -2682288000 // number of seconds between the unix epoch and the epoch for time.Time.wall (1 jan 1885)
+)
+
+// formatTime writes formatted value of a time.Time to v.Value.
+// See $GOROOT/src/time/time.go for a description of time.Time internals.
+func (v *Variable) formatTime() {
+	wallv := v.fieldVariable("wall")
+	extv := v.fieldVariable("ext")
+	if wallv == nil || extv == nil || wallv.Unreadable != nil || extv.Unreadable != nil || wallv.Value == nil || extv.Value == nil {
+		return
+	}
+
+	var loc *time.Location
+
+	locv := v.fieldVariable("loc")
+	if locv != nil && locv.Unreadable == nil {
+		namev := locv.loadFieldNamed("name")
+		if namev != nil && namev.Unreadable == nil {
+			name := constant.StringVal(namev.Value)
+			loc, _ = time.LoadLocation(name)
+		}
+	}
+
+	wall, _ := constant.Uint64Val(wallv.Value)
+	ext, _ := constant.Int64Val(extv.Value)
+
+	hasMonotonic := (wall & timeTimeWallHasMonotonicBit) != 0
+	if hasMonotonic {
+		// the 33-bit field of wall holds a 33-bit unsigned wall
+		// seconds since Jan 1 year 1885, and ext holds a signed 64-bit monotonic
+		// clock reading, nanoseconds since process start
+		sec := int64(wall << 1 >> (wallNsecShift + 1)) // seconds since 1 Jan 1885
+		t := time.Unix(sec+unixTimestampOfWallEpoch, 0).UTC()
+		if loc != nil {
+			t = t.In(loc)
+		}
+		v.Value = constant.MakeString(fmt.Sprintf("%s, %+d", t.Format(time.RFC3339), ext))
+	} else {
+		// the full signed 64-bit wall seconds since Jan 1 year 1 is stored in ext
+		var t time.Time
+		for ext > int64(maxAddSeconds/time.Second) {
+			t = t.Add(maxAddSeconds)
+			ext -= int64(maxAddSeconds / time.Second)
+		}
+		t = t.Add(time.Duration(ext) * time.Second)
+		if loc != nil {
+			t = t.In(loc)
+		}
+		v.Value = constant.MakeString(t.Format(time.RFC3339))
+	}
+}
